from django import forms
from django.contrib.gis.measure import D


class DistanceField(forms.Field):

    def to_python(self, value):
        if value in self.empty_values:
            return None
        return (value['geometry'], D(**{value['unit']: value['value']}))



from collections import OrderedDict

import graphene
from graphene.types.generic import GenericScalar
from graphene_django.types import DjangoObjectType, DjangoObjectTypeOptions

from .. import resolver

__all__ = ['GeoJSONType']


class OrderedFields(OrderedDict):

    def update(self, other=None, **kwargs):
        if other is not None:
            global_id_field = other.pop('id', None)

            if global_id_field is not None:
                self['id'] = global_id_field

            Properties = type('Properties', (
                self['properties']._type,
            ), other)

            self['properties'] = graphene.Field(Properties)
        else:
            super().update(**kwargs)


class GeoJSONTypeOptions(DjangoObjectTypeOptions):
    geojson_field = None

    def __setattr__(self, name, value):
        if name == 'fields':
            geometry_field = value.pop(self.geojson_field, None)

            assert geometry_field is not None, (
                'Unrecognized field `{}`'.format(self.geojson_field)
            )

            geometry_field.name = 'geometry'
            bbox_field = graphene.Field(
                GenericScalar,
                default_value=self.geojson_field)

            primary_key = self.model._meta.pk.name
            primary_key_field = value.pop(primary_key, None)

            value.update({
                key: attr for key, attr in self.class_type.__dict__.items()
                if key.startswith('resolve') and callable(attr)
            })

            Properties = type('Properties', (graphene.ObjectType,), value)

            fields = [
                ('type', graphene.Field(graphene.String)),
                (self.geojson_field, geometry_field),
                ('bbox', bbox_field),
                ('properties', graphene.Field(Properties)),
            ]

            if primary_key_field is not None:
                fields.insert(1, (primary_key, primary_key_field))

            value = OrderedFields(fields)

        super().__setattr__(name, value)


class GeoJSONType(DjangoObjectType):

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, name=None, _meta=None,
                                    geojson_field=None, **options):

        assert geojson_field is not None, (
            '`{}` should either include a `Meta.geojson_field`'
            ' attribute.'.format(name or cls.__name__)
        )

        if _meta is None:
            _meta = GeoJSONTypeOptions(cls)
            _meta.geojson_field = geojson_field

        options.setdefault('default_resolver', resolver.feature_resolver)
        super().__init_subclass_with_meta__(name=name, _meta=_meta,  **options)



from .distance import Distance
from .feature import GeoJSONType
from .geometry import Geometry, GeometryObjectType


__all__ = [
    'Distance',
    'Geometry',
    'GeometryObjectType',
    'GeoJSONType',
]



import json

from django.contrib.gis.geos import GEOSGeometry

import graphene
from graphene.types.generic import GenericScalar
from graphql.language import ast

from .. import resolver

__all__ = [
    'Geometry',
    'GeometryObjectType',
]


class Geometry(graphene.Scalar):

    @classmethod
    def serialize(cls, value):
        return json.loads(value.geojson)

    @classmethod
    def parse_literal(cls, node):
        if isinstance(node, ast.StringValue):
            return cls.parse_value(node.value)
        return None

    @classmethod
    def parse_value(cls, value):
        if isinstance(value, dict):
            value = json.dumps(value)
        return GEOSGeometry(value)


class GeometryObjectType(graphene.ObjectType):
    type = graphene.String()
    coordinates = GenericScalar()

    class Meta:
        default_resolver = resolver.geometry_resolver



from django.contrib.gis import forms
from django.contrib.gis.db import models

import django_filters

from . import fields


class GeometryFilter(django_filters.Filter):
    field_class = forms.GeometryField


class DistanceFilter(django_filters.Filter):
    field_class = fields.DistanceField


FILTER_DEFAULTS = django_filters.FilterSet.FILTER_DEFAULTS
FILTER_DEFAULTS.update({
    models.GeometryField: {
        'filter_class': GeometryFilter,
    },
})


class GeometryFilterSet(django_filters.FilterSet):
    FILTER_DEFAULTS = FILTER_DEFAULTS

    @classmethod
    def filter_for_lookup(cls, field, lookup_type):
        if lookup_type.startswith('distance') or lookup_type == 'dwithin':
            return DistanceFilter, {}
        return super().filter_for_lookup(field, lookup_type)



from django.contrib.gis.measure import D

import graphene

from .geometry import Geometry

__all__ = ['Distance']


DistanceUnitEnum = graphene.Enum('DistanceUnitEnum', [
    (value, value) for value in D.UNITS.keys()
])


class Distance(graphene.InputObjectType):
    unit = DistanceUnitEnum(required=True)
    value = graphene.Float(required=True)
    geometry = Geometry(required=True)



from django.contrib.gis import forms
from django.contrib.gis.db import models

import graphene
from graphene_django.converter import convert_django_field
from graphene_django.form_converter import convert_form_field

from . import fields, types


@convert_django_field.register(models.GeometryField)
def convert_field_to_geometry(field, registry=None):
    return graphene.Field(
        types.GeometryObjectType,
        description=field.help_text,
        required=not field.null)


@convert_form_field.register(forms.GeometryField)
def convert_form_field_to_geometry(field):
    return types.Geometry(
        description=field.help_text,
        required=field.required)


@convert_form_field.register(fields.DistanceField)
def convert_form_field_to_distance(field):
    return types.Distance(
        description=field.help_text,
        required=field.required)



import json


def geometry_resolver(attname, default_value, root, info, **args):
    if default_value is not None:
        root = root or default_value
    return json.loads(root.geojson)[attname]


def feature_resolver(attname, default_value, root, info, **args):
    if attname == 'type':
        return 'Feature'

    elif info.field_name == 'geometry':
        return getattr(root, attname)

    elif attname == 'bbox':
        geometry = getattr(root, default_value)

        if geometry is not None:
            return geometry.extent
        return None

    return root
